#!/usr/bin/perl


use strict;
use warnings;

package      MPEGSTREAM;
require      Exporter;

our @ISA       = qw(Exporter);
our @EXPORT    = qw(	add_PES_packet
			bitrate
			streamtype
			streamcode
			print_stats
			output_files
			process_PES_packets
		   );

our $VERSION   = 0.01;		        # Version number


# FIXME handle PTS timer overflows

sub new {
    my $invocant = shift;
    my $class   = ref($invocant) || $invocant;
    my $self = {
        streamcode  	=> "PES_stream", #will be overridden
        streamtype  	=> undef, 	 # not used yet
        bitrate     	=> undef,	# only for Audio
	freq		=> undef,	# only for Audio 
	copyright	=> undef,	# only for Audio
	frame_bytes	=> undef,	# only for Audio
	frame_ticks	=> undef,	# one tick is 1 / 90000 of a second
	frameno     	=> 0,		# READ frames counter, not written frames
	frames_in_GOP	=> 0,		# only used for video, needed because only ts at GOP starts are stored
	GOPno		=> 0,		# just for video stats
	cut_flag	=> 0,		# if a cut is detected in this stream, the flag is set to tell the parent
	packet_start	=> undef,	# a signature that is found a the beginning of a frame, should read frame_start
	packet_rest	=> "",		# only audio, if a frame does not finish in a PES-Packet, the first part of the frame is stored here
	received_packets=> 0,		# used for stats (not yet)
	payload		=> undef,	# a place to store the filehandle 
	check_frames    => undef,	# a placeholder for the sub that actually manages the frames of a given type (mp2 / ac3 / mpv)
	masterstream	=> "",		# here the stream_id of the videostream is stored (it is not always e0 as I thought first)
	outputpath	=> "./",	# where to store the result files
	inserted_packets=> 0,		# packets that are inserted at a cut start or end, not sure whether it is used anymore
	cutcounter	=> 0,		# used as an index to the cuts list (see below), for sync we need the current and the previous cut
	chunk_start_correction => 0,	# a string that stores info about how many frames have to be inserted where to sync the start of an audio chunk
	chunk_end_correction => 0,	# a string that stores info about how many frames have to be inserted where to sync the end of an audio chunk
	frames_since_last_flush => 0,	# counts audio frames in buffer, if it is too big, frames are flushed to disk
	frame_buffer	=> [],		# an array that holds the buffered audio frames
	first_frame_in_buffer => 0,	# the first audio frame in the buffer is actualy the nth frame read. n is stored here
	frames_written	=> 0,		# used only for stats
	last_frame_w_ts	=> undef, 	#needed to  compute  cut and sync correction, maybe not anymore
	timestamps	=> [],		# all timesstamps are stored here, probably a waste of memory, but maybe used for stricter checks
	cuts		=> [],		# an array of scalars that contain vital information about cuts
	sync_frames	=> [ 0 ],	# number of frames that were inserted to sync, not to sync the end of a cunk
	final_desync	=> [ 0 ], 	# how many ticks desync after correction? Used in the next sync attempt
	stop		=> 0,		# indicates whther the script should stop often to allow you to read debug output

	@_,				# Override previous attributes
	};
	if ($self->{streamcode} eq "PES_stream")
		{
		
		if (scalar(@{$self->{files}}) == 0)
			{
			print "No input files specified, exiting\n"; exit;
			}
		
		foreach (@{$self->{files}})  # check whether all files that we got as parameters exist
			{
			if (!(-e $_)) {print "Input file $_ not found\n"; exit}
			}
		$self->{packets_before_sync} = 0;
		print "Stop Flag found, script will stop frequently until you press enter\n" if $self->{stop};
		print "dump only the first $self->{dump_packets} packets\n"  if $self->{dump_packets};
		foreach (keys (%{$self->{ignore_streams}})) {print "Ignoreing stream $_\n";}
		print "Lots of debug stuff will be printed\n"  if $self->{debug};
		
		init_PES_stream($self);
		}
		
	return bless $self, $class;
}

sub add_PES_packet
	{
	my $self = shift;
	my $PES_packet = ${shift @_};
	#print "In add_PES_packet for stream $self->{streamcode}\n" if $self->{debug};
	#print "Received PES packet of length " . length($PES_packet) . "\n" if $self->{debug};
	my %pes_header = analyse_PES_header(\$PES_packet);
	#if ($self->{debug})
	#	{
	#	foreach (keys(%pes_header))
	#		{
	#		print "Header analysis returned $_: $pes_header{$_}\n";
	#		}
	#	}
	$self->{received_packets}++;
	print "Received PES packet number $self->{received_packets} (for stream $self->{streamcode}) with a length of " . length($PES_packet) . "\n" if $self->{debug};
	if ($pes_header{PTS_DTS_flags} ne "00")
		{
		push @{$self->{timestamps}}, $pes_header{PTS_decimal};
		if (scalar(@{$self->{timestamps}}) > 1)
			{
			my $last = ${$self->{timestamps}}[-1]; 
			my $beforelast = ${$self->{timestamps}}[-2]; 
			if ($last - $beforelast > 90000) # THIS SHOULD BE 63000 BUT MANY STREAMS HAVE LONGER INTERVALLS
				{
				$self->{cutcounter}++;
				print "Cut detected in stream $self->{streamcode} at ". ($last/90000) ."\n";
				print "Storing cut info in array member $self->{cutcounter}\n" if ($self->{debug});
				print "Diff between ts: " . ($last - $beforelast) . "\n"if ($self->{debug});
				my $dummy = <STDIN> if ($self->{stop});
				
			# Info about the cuts store in a : separated string: first_frame, first_ts, last_ts_bef_cut, 
			#                                                                                          last_frame_w_ts_bef_cut, stream_code			
				$self->{cuts}[$self->{cutcounter}] .= $self->{frameno} . " : ";
				$self->{cuts}[$self->{cutcounter}] .= $last . " : ";
				$self->{cuts}[$self->{cutcounter}] .= $beforelast  . " : ";
				$self->{cuts}[$self->{cutcounter}] .= $self->{last_frame_w_ts} . " : ";
				$self->{cuts}[$self->{cutcounter}] .= $self->{streamcode} . " : ";
				print "Master videostream set to $self->{masterstream}\n" if $self->{debug};
				if ($self->{streamcode} eq $self->{masterstream})
					{
					$self->{cuts}[$self->{cutcounter}] .= $self->{frames_in_last_GOP} . " : ";
					}
				
				print "cut detected at frame $self->{frameno} in stream $self->{streamcode}\n" if $self->{debug};
				$self->{cut_flag} = 1;
				}
			}
		$self->{last_frame_w_ts} = $self->{frameno};
		}
	if (!($self->{check_frames}))
		{
		init_stream($self, substr($PES_packet, $pes_header{payload_start}), \%pes_header);
		if (!($self->{check_frames}))
			{
			print "no sync with the stream $self->{streamcode} yet\n"; return;
			}
		}

		&{$self->{check_frames}} ($self, substr($PES_packet,$pes_header{payload_start}), \%pes_header)	;
	}

sub init_stream
	{
	my $self = shift;
	my $payload = shift;
	my %pes_header = %{shift @_};
	my $decimal_code = hex($pes_header{stream_id});
	print "we got the decimal code $decimal_code for $pes_header{stream_id}\n"if $self->{debug};
	if (($decimal_code > 191) && ($decimal_code < 224))
		{
		print "we got the stream $decimal_code ($pes_header{stream_id}), checking Audio\n"if $self->{debug};
		my $dummy = <STDIN> if $self->{stop};
		analyse_audio_stream($self, $payload, \%pes_header);
		#$self->{streamtype} = "audio";
		}
	elsif (($decimal_code > 223) && ($decimal_code < 240))	
		{
		print "we got the stream $decimal_code ($pes_header{stream_id}), checking Video\n"if $self->{debug};
		my $dummy = <STDIN> if $self->{stop};
		analyse_video_stream($self, $payload, \%pes_header);
		$self->{streamtype} = "video";
		}
	elsif ($decimal_code == 189)
		{
		analyse_ac3_stream($self, $payload, \%pes_header);
		$self->{streamtype} = "AC3 audio";
		}
	else	
		{
		print "unknown stream type found, skipping contents of stream $pes_header{stream_id}\n";
		#$self->{check_frames} = \&do_nothing;
		}
	}

sub return_cuts
	{
	my $self = shift;
	return $self->{cuts};
	}

sub output_files
	{
	my $self = shift;
	my @filelist;
	if ($self->{streamcode} ne "PES_stream") {return}
	#foreach (keys(%{$self->{streams}}))
	#	{
	#	print "We have the streams $_\n";
	#	}
	print "$self->{streams}{$self->{masterstream}}{outputpath}/$self->{masterstream}.mpv\n" if $self->{debug};
	push @filelist,"$self->{streams}{$self->{masterstream}}{outputpath}/$self->{masterstream}.mpv";
	foreach (keys (%{$self->{streams}}))
		{
		if ($self->{streams}{$_}{streamcode} eq $self->{masterstream}) {next};
		if ($self->{ignore_streams}{$_})
			{
			print "ignoring packet for stream $_ \n" if $self->{debug};
			next;
			}
		if ($self->{streams}{$_}{streamcode} =~/^c/)
			{
			push @filelist,"$self->{streams}{$_}{outputpath}/$_.mpa";
			print "$self->{streams}{$_}{outputpath}/$_.mpa\n" if $self->{debug};
			}
		if ($self->{streams}{$_}{streamcode} eq "bd")
			{
			push @filelist,"$self->{streams}{$_}{outputpath}/$_.ac3";
			}
		
		}
	return \@filelist;
	}



sub compute_desync
	{
	my $self = shift;
	
	
	my $current_cut = $self->{cutcounter};
	my @mastertime = @{$self->{streams}{$self->{masterstream}}->return_cuts()} if (! $self->{dump_payload});
	
	foreach (@mastertime)
		{
		print "For the video stream we have the following cuts: $_\n" if ($self->{debug});
		}
	foreach (keys (%{$self->{streams}}))
		{
		if ($_ eq $self->{masterstream}) {next}
		if ($self->{streams}{$_}->{dump_payload}) 
			{
			print "Dumping stream $_ without sync\n" if $self->{debug};
			next;
			}
		if ($self->{ignore_streams}{$_})
			{
			print "ignoring packet for stream $_ \n" if $self->{debug};
			next;
			}
		if ($self->{cutcounter} != $self->{streams}{$_}->{cutcounter})
			{
			print "Whoa, cut that has not been detected in one stream...\n";
			$self->{packets_before_sync} = 100;
			return;
			}
		if ($current_cut > 0) # we are not at the beginning of the file and have to correct the end of the prev chunk
			{
			print "We should sync stream $_\n" if($self->{debug});
			$self->{streams}{$_}->sync_chunk_end(\@mastertime);
			}
		my $success = $self->{streams}{$_}->sync_chunk_start( \@mastertime);
 		if ($success == -1)
 			{
 			$self->{ignore_streams}{$_} = 1;
 			print "Timing of stream $_ was totally messed, ignoring from now on...\n";
			}
				
		$self->{streams}{$_}->flush_audio_frames($self->{streams}{$_});
		}
	$self->{cutcounter}++;
	}
	
sub sync_chunk_start
	{
	my $self = shift;
	
	print "trying to sync start of $self->{streamcode} \n" if ($self->{debug});
	my @mastertime = @{shift @_};
	
	foreach (@mastertime)
		{
		print "mastertime in sync start: $_\n" if ($self->{debug});
		}
	
	
	my $current_cut = $self->{cutcounter};
	print "current cut is $current_cut\n" if ($self->{debug});
	my @audiotime = @{$self->return_cuts()};
	foreach (@audiotime)
		{
		print "audiotime in sync start: $_\n" if ($self->{debug});
		}
	
	
	my $last_desync = $self->{final_desync}[$current_cut];
	
	# We stored a bit of info about the cuts in a single string that we split back into scalars here
	my  ($v_start_frame, $v_start_time, $prev_v_end_time, $prev_v_end_frame, $v_stream_code, $frames_in_last_GOP)
		= split / : /,$mastertime[$current_cut];

	

	my  ($a_start_frame, $a_start_time, $prev_a_end_time, $prev_a_end_frame, $a_stream_code) 
		= split / : /,$audiotime[$current_cut];
	
		
	my $av_shift = $a_start_time - $v_start_time; # Only for this chunk!!! Assumes that the first Audio and Video frames in this chunk have a ts
	print "AV shift before adding last desync $av_shift\n" if ($self->{debug});
	$av_shift += $last_desync;
	print "The first Audio and Video frames (v-a) differ by $av_shift ticks\n" if ($self->{debug});
	my $add_frames_start = int($av_shift / $self->{frame_ticks});
	if ($add_frames_start >= 1000) # Timing is totaly mixed up
 		{
 		$add_frames_start = 0;
 		return -1;
 		}
	my $final_desync = $av_shift  - ($add_frames_start * $self->{frame_ticks});
	print "Correcting by adding $add_frames_start at $a_start_frame, $final_desync not corrected\n" if ($self->{debug}); 
	

	$self->{inserted_packets} += abs($add_frames_start);
	$self->{sync_frames}[$current_cut] = $add_frames_start;
	$self->{final_desync}[$current_cut] = $final_desync;
	$self->{chunk_start_correction} = "$add_frames_start : "  . ($a_start_frame);
	return 0;
	}
	
sub sync_chunk_end
	{
	my $self = shift;
	my @mastertime = @{shift @_};
	
	my $current_cut = $self->{cutcounter};
	
	my @audiotime = @{$self->return_cuts()};
	print "trying to sync end of $self->{streamcode}  at cut $current_cut \n" if ($self->{debug});
	# We stored a bit of info about the cuts in a single string that we split back into scalars here
	my  ($v_start_frame, $v_start_time, $prev_v_end_time, $prev_v_end_frame, $v_stream_code, $frames_in_last_GOP)
		= split / : /,$mastertime[$current_cut];

	my  ($a_start_frame, $a_start_time, $prev_a_end_time, $prev_a_end_frame, $a_stream_code) 
		= split / : /,$audiotime[$current_cut];
	
	my $add_frames_start = ${$self->{sync_frames}}[$current_cut-1];
	my $start_desync = ${$self->{final_desync}}[$current_cut-1];
	
	# to be able to sync the end of a chnuk we have to get information about the chunk that we sync, and the prev chunk
	# so we split the string with the infos from the prev chunk also back in scalars
	my  ($prev_v_start_frame, $prev_v_start_time, $prev_prev_v_end_time, $prev_prev_v_end_frame, $prev_v_stream_code)
	= split / : /,$mastertime[$current_cut-1];
	my  ($prev_a_start_frame, $prev_a_start_time, $prev_prev_a_end_time, $prev_prev_a_end_frame, $prev_a_stream_code) 
	= split / : /,$audiotime[$current_cut-1];
	
	my $v_chunk_frames = $prev_v_end_frame - $prev_v_start_frame ;
	# Here we try to compensate for frames that do not have a ts at the end of the chunk, they do also contribute to the length of the chunk ...
	my $a_chunk_frames = $prev_a_end_frame - $prev_a_start_frame + ($a_start_frame - $prev_a_end_frame)+ $add_frames_start;
	
	# Since only the ts of GOPas starts are stored, we have to add the number of frames in the last GOP to get the real length
	my $v_chunk_time = $prev_v_end_time - $prev_v_start_time + (($frames_in_last_GOP - 1 )  * 3600);
	
	# Here we have to be even more careful: we synced the start of the chunk by addition/deletion of some frames, so we need to add them
	# Then we have the same problem as above, potential audio frames without ts at the end
	my $a_chunk_time = $prev_a_end_time - $prev_a_start_time + (($a_start_frame + $add_frames_start - $prev_a_end_frame) * $self->{frame_ticks}) ;
	
	my $v_diff = $v_chunk_time - ($v_chunk_frames * 3600);
	my $a_diff = $a_chunk_time - ($a_chunk_frames * $self->{frame_ticks});
	print "astartframe: $a_start_frame, prev_a_end_frame: $prev_a_end_frame, add_frames_start: $add_frames_start\n" if ($self->{debug});
	print "v_chunk_frames: $v_chunk_frames, a_chunk_frames: $a_chunk_frames\n" if ($self->{debug});
	print "v_chunk_time: $v_chunk_time, a_chunk_time: $a_chunk_time\n" if ($self->{debug});
	print "current frame is $self->{frameno}\n" if ($self->{debug});
	print "prev_v_start_time: $prev_v_start_time \n" if ($self->{debug});
	print "last frame in prev chunk was $prev_a_end_frame\n" if ($self->{debug});
	print "Found a video diff (frames vs ts) of $v_diff for stream $self->{streamcode}\n" if ($self->{debug});	
	print "Found a audio diff (frames vs ts) of $a_diff for stream $self->{streamcode}\n\n" if ($self->{debug});
	my $dummy = <STDIN> if ($self->{stop});
	my $av_length_diff = $v_chunk_time - $a_chunk_time + $start_desync;
	my $add_frames_end = int($av_length_diff / $self->{frame_ticks});
	my $final_length_diff = $av_length_diff  -  ($add_frames_end * $self->{frame_ticks});
	$self->{final_desync}[$current_cut] = $final_length_diff;
	
	$self->{chunk_end_correction} = "$add_frames_end : "  . ($a_start_frame - 1) ;
	
	$dummy = <STDIN> if $self->{stop};
	# Some helpful info for checking whats going on ...
	#print "In absolute packets that is  $self->{inserted_packets}\n";
	#print "in cut " . ($current_cut) . " we should have a final_desync of " . $self->{final_desync}[$current_cut] . "\n";
	#print "Some statistics about the previous video chunk:\n";
	#print "prev_v_end_frame $prev_v_end_frame  prev_v_start_frame $prev_v_start_frame\n";
	#print "v_start_frame, $v_start_frame,  v_start_time  $v_start_time\n";
	#print "v-length in ticks: $v_chunk_time in frames: $v_chunk_frames\n";
	#print "a-length in ticks: $a_chunk_time in frames: $a_chunk_frames\n";
	#print "diff of length (ts vs frame * frametime): v: $v_diff a: $a_diff\n";
	#print "av length diff: $av_length_diff\n";
	#print "correcting by adding $add_frames_end to the audio, final length diff: $final_length_diff\n";
	#print "Takining start_desync into account\n" . "X" x 20 . "\n";
	}
	
sub analyse_ac3_stream
	{
	
	# lots of boring typing here ....
	my %AC3_framelength =
	(
	"32000" => {
		"000000"	=> "192",
		"000001"	=> "192",
		"000010"	=> "240",
		"000011"	=> "240",
		"000100"	=> "288",
		"000101"	=> "288",
		"000110"	=> "336",
		"000111"	=> "336",
		"001000"	=> "384",
		"001001"	=> "384",
		"001010"	=> "480",
		"001011"	=> "480",
		"001100"	=> "576",
		"001101"	=> "576",
		"001110"	=> "672",
		"001111"	=> "672",
		"010000"	=> "768",
		"010001"	=> "768",
		"010010"	=> "960",
		"010011"	=> "960",
		"010100"	=> "1152",
		"010101"	=> "1152",
		"010110"	=> "1344",
		"010111"	=> "1344",
		"011000"	=> "1536",
		"011001"	=> "1536",
		"011010"	=> "1920",
		"011011"	=> "1920",
		"011100"	=> "2304",
		"011101"	=> "2304",
		"011110"	=> "2688",
		"011111"	=> "2688",
		"100000"	=> "3072",
		"100001"	=> "3072",
		"100010"	=> "3456",
		"100011"	=> "3456",
		"100100"	=> "3840",
		"100101"	=> "3840",
		},
	"44100" => {
		"000000"	=> "138",
		"000001"	=> "140",
		"000010"	=> "194",
		"000011"	=> "196",
		"000100"	=> "208",
		"000101"	=> "210",
		"000110"	=> "242",
		"000111"	=> "244",
		"001000"	=> "278",
		"001001"	=> "280",
		"001010"	=> "348",
		"001011"	=> "350",
		"001100"	=> "416",
		"001101"	=> "418",
		"001110"	=> "486",
		"001111"	=> "488",
		"010000"	=> "556",
		"010001"	=> "558",
		"010010"	=> "696",
		"010011"	=> "698",
		"010100"	=> "834",
		"010101"	=> "836",
		"010110"	=> "974",
		"010111"	=> "976",
		"011000"	=> "1114",
		"011001"	=> "1116",
		"011010"	=> "1392",
		"011011"	=> "1394",
		"011100"	=> "1670",
		"011101"	=> "1672",
		"011110"	=> "1950",
		"011111"	=> "1952",
		"100000"	=> "2228",
		"100001"	=> "2230",
		"100010"	=> "2506",
		"100011"	=> "2508",
		"100100"	=> "2786",
		"100101"	=> "2788",
		},
	"48000" => {
		"000000"	=> "128",
		"000001"	=> "120",
		"000010"	=> "160",
		"000011"	=> "160",
		"000100"	=> "192",
		"000101"	=> "192",
		"000110"	=> "224",
		"000111"	=> "224",
		"001000"	=> "256",
		"001001"	=> "256",
		"001010"	=> "320",
		"001011"	=> "320",
		"001100"	=> "384",
		"001101"	=> "384",
		"001110"	=> "448",
		"001111"	=> "448",
		"010000"	=> "512",
		"010001"	=> "512",
		"010010"	=> "640",
		"010011"	=> "640",
		"010100"	=> "768",
		"010101"	=> "768",
		"010110"	=> "896",
		"010111"	=> "896",
		"011000"	=> "1024",
		"011001"	=> "1024",
		"011010"	=> "1280",
		"011011"	=> "1280",
		"011100"	=> "1336",
		"011101"	=> "1336",
		"011110"	=> "1792",
		"011111"	=> "1792",
		"100000"	=> "2048",
		"100001"	=> "2048",
		"100010"	=> "2304",
		"100011"	=> "2304",
		"100100"	=> "2560",
		"100101"	=> "2560",
		},
	"bitrate" => {
		"000000"	=> "32000",
		"000001"	=> "32000",
		"000010"	=> "40000",
		"000011"	=> "40000",
		"000100"	=> "48000",
		"000101"	=> "48000",
		"000110"	=> "56000",
		"000111"	=> "56000",
		"001000"	=> "64000",
		"001001"	=> "64000",
		"001010"	=> "80000",
		"001011"	=> "80000",
		"001100"	=> "96000",
		"001101"	=> "96000",
		"001110"	=> "112000",
		"001111"	=> "112000",
		"010000"	=> "128000",
		"010001"	=> "128000",
		"010010"	=> "160000",
		"010011"	=> "160000",
		"010100"	=> "192000",
		"010101"	=> "192000",
		"010110"	=> "224000",
		"010111"	=> "224000",
		"011000"	=> "256000",
		"011001"	=> "256000",
		"011010"	=> "320000",
		"011011"	=> "320000",
		"011100"	=> "384000",
		"011101"	=> "384000",
		"011110"	=> "448000",
		"011111"	=> "448000",
		"100000"	=> "512000",
		"100001"	=> "512000",
		"100010"	=> "576000",
		"100011"	=> "576000",
		"100100"	=> "640000",
		"100101"	=> "640000",
		},
	);
		
	
	my $self = shift;
	my $payload = shift;
	my %pes_header = %{shift @_};
	if (!($pes_header{PTS_decimal})) {print "Don't accept first packet of a stream w/o ts\n";return}
	
	my $bits = unpack("B64", substr($payload, 0, 8));
	#for (my $i = 0; $i < 64; $i += 8)
	#	{
	#	print "bits " . ($i +1 ) . " to " . ($i + 8) ."\n";
	#	print substr($bits, $i, 8) . "\n";
	#	}
	if (substr($bits,0,16) ne "0000101101110111")
		{
		print "No Audio syncword at begin, better luck next packet\n" if $self->{debug};
		print substr($bits,0,16) . " found, 0000101101110111 expected\n" if $self->{debug};;
		my $dummy = <STDIN> if $self->{stop};
		return;
		}
	
	
	$self->{packet_start} = substr($payload,0,2);
	my $fscod = substr($bits,32,2);
	
	if ($fscod == "00")  {$self->{freq} = 48000}
	elsif ($fscod == "01")  {$self->{freq} = 44100}
	elsif ($fscod == "11")  {$self->{freq} = 32000}
	elsif ($fscod == "10")  {die "Illeagal AC3 freq\n"}
	
	
	my $frmsizecod = substr($bits,34,6);
	#my $bsinfo = substr($bits, 40, 5);
	#my $bsmod = substr($bits, 45, 3);
	$self->{frame_bytes} = $AC3_framelength{$self->{freq}}{$frmsizecod};
	$self->{bitrate} = $AC3_framelength{bitrate}{$frmsizecod};
	
	my $acmod = substr($bits, 48, 3);
	if    ($acmod == "000") {$self->{mode} = "1+1"}
	elsif ($acmod == "001") {$self->{mode} = "1/0"}
	elsif ($acmod == "010") {$self->{mode} = "2/0"}
	elsif ($acmod == "011") {$self->{mode} = "3/0"}
	elsif ($acmod == "100") {$self->{mode} = "2/1"}
	elsif ($acmod == "101") {$self->{mode} = "3/1"}
	elsif ($acmod == "110") {$self->{mode} = "2/2"}
	elsif ($acmod == "111") {$self->{mode} = "3/2"}
	
	##################################################################
	############ ONE FRAME is 32 ms
	##################################################################
	# The following is not correct, we need some more logic since fields depend on the presence of flags before
	#my $cmixlevel = substr($bits, 19, 2);
	#my $surmixlevel = substr($bits, 21, 2);
	#my $dsurmod = substr($bits, 23, 2);
	#if ($dsurmod == "00")    {$self->{dolby} = "?"}
	#elsif ($dsurmod == "01") {$self->{dolby} = "no dolby surround"}
	#elsif ($dsurmod == "10") {$self->{dolby} = "dolby surround"}
	#elsif ($dsurmod == "11") {$self->{dolby} = "should not stand here"}
	#my $lfeon = substr($bits, 25, 1);
	#$self->{subwoofer} = $lfeon;
	#my $dialnorm = substr($bits, 26, 5);
	#my $compre = substr($bits, 31, 1);
	#my $compr  = substr($bits, 32, 8) if ($compre);
	if ($fscod ne "00") {print "not an 48 KHz stream, exiting\n"; exit}
	$self->{frame_ticks} =  8 * $self->{frame_bytes} * 90000 / $self->{bitrate};
		
	print "framesizecode is $frmsizecod and fscod is $fscod\n" if $self->{debug};
	print "payload length is " . length($payload) . "\n" if $self->{debug};
	$self->{cuts}[0] = "$self->{frameno}  : $pes_header{PTS_decimal} : -1 : -1 : $self->{streamcode} : ";
	print "First info from bd $self->{frameno}  : $pes_header{PTS_decimal} : -1 : -1 : $self->{streamcode} : \n" if $self->{debug};
	$self->{cut_flag} = 1;
	my $dummy = <STDIN> if $self->{stop};
	print "Some info about the AC3-Stream:\n bitrate: $self->{bitrate}\n Sample-Frequency: $self->{freq}\n" if $self->{debug};
	print " frmsizecod: $frmsizecod\n Audio-Mode: $self->{mode}\n Bytes per frame: $self->{frame_bytes}\n" if $self->{debug};
	print " frame ticks: $self->{frame_ticks}\n" if $self->{debug};
	
	open $self->{outfile}, ">$self->{outputpath}/$self->{streamcode}.ac3" || die "Can not open file for ac3: $!\n" if (!($self->{outfile}));
	$self->{check_frames} = \&check_AC3_frames;
	}

sub check_mp2_frames
	{
	my $self = shift;
	my $payload = shift;
	my %pes_header = %{shift @_};
	my $counter = 0;
	my $FH = $self->{outfile};
	
	if ($self->{dump_payload})
			{
			print "Dumping payload of stream $self->{streamcode}, ($self->{dump_payload} to dump left)\n";
			my $DUMPFH;
			if (! $self->{"$self->{streamcode}.dump"})
				{
				print "Trying to open dumpfile\n";
				open $DUMPFH, ">$self->{streamcode}.dump" || die "Can not open dumpfile: $!\n";
				$self->{"$self->{streamcode}.dump"} = $DUMPFH;
				print "opened dumpfile\n";
				}
			$DUMPFH = $self->{"$self->{streamcode}.dump"};
			print $DUMPFH $payload;
			if ($self->{dump_payload}-- == 1)
				{
				exit;
				}
			return;
			}
	
	
	$payload = $self->{packet_rest} . $payload;
	while (($counter+1)*$self->{frame_bytes} <= (length($payload)))
		{
		if ($self->{packet_start} eq (substr($payload,($counter*$self->{frame_bytes}),2)))
			{
			push @{$self->{frame_buffer}}, substr($payload,($counter * $self->{frame_bytes}), $self->{frame_bytes});
			if (!($self->{single_frame}))
				{
				$self->{single_frame} = substr($payload,($counter * $self->{frame_bytes}), $self->{frame_bytes});
				print "Defined first audio frame as single frame for later sync in stream $self->{streamcode}\n" if $self->{debug};
				print "Frame has a length of " . length($self->{single_frame}) . "\n" if $self->{debug};
				my $dummy = <STDIN> if ($self->{stop});
				}
			$counter++;
			$self->{frameno}++;
			$self->{frames_since_last_flush}++;
			}
		else 	{
			search_audio_sync($self, $payload, \%pes_header);
			print "searching audio sync...\n" if $self->{debug};
			return;
			}
		}
	if ($counter*$self->{frame_bytes} < (length($payload)))
		{
		$self->{packet_rest} = substr($payload,($counter * $self->{frame_bytes}));
		}
	else	
		{ 
		$self->{packet_rest} ="";
		}
	}
sub check_AC3_frames
	{
	my $self = shift;
	my $payload = shift;
	my %pes_header = %{shift @_};
	my $counter = 0;
	my $FH = $self->{outfile};
	
	$payload = $self->{packet_rest} . $payload if (!($pes_header{data_align}));
	print "we got a packet rest of " . (length($self->{packet_rest})) . "\n" if $self->{debug};

	while (($counter+1)*$self->{frame_bytes} <= (length($payload)))
		{
		if ($self->{packet_start} eq (substr($payload,($counter*$self->{frame_bytes}),2)))
			{
			push @{$self->{frame_buffer}}, substr($payload,($counter * $self->{frame_bytes}), $self->{frame_bytes});
			if (!($self->{single_frame}))
				{
				$self->{single_frame} = substr($payload,($counter * $self->{frame_bytes}), $self->{frame_bytes});
				print "Defined first audio frame as single frame for later sync in stream $self->{streamcode}\n" if $self->{debug};
				print "Frame has a length of " . length($self->{single_frame}) . "\n" if $self->{debug};
				my $dummy = <STDIN> if ($self->{stop});
				}
			$counter++;
			$self->{frameno}++;
			$self->{frames_since_last_flush}++;
			}
		else 	{
			print "Data align bit is: $pes_header{data_align}\n" if $self->{debug};
			print "$self->{frame_bytes} bytes is the packet length\n" if $self->{debug};
			my $should_start = unpack ("H4", $self->{packet_start});
			my $is_start = unpack ("H4", substr($payload,($counter*$self->{frame_bytes}),2));
			my $trans = unpack ("H12", $payload);
			print "$is_start is the start\ntrans is $trans\n" if $self->{debug};
			print "$should_start should be the start \n" if $self->{debug};
			print "no audio sync for AC3 found, skipping packet\n" if $self->{debug};
			$self->{packet_rest} ="";
			my $dummy = <STDIN> if $self->{stop};
			return;
			}
		}
	if ($counter*$self->{frame_bytes} < (length($payload)))
		{
		#print "we have a packet rest for bd???\n";
		
		$self->{packet_rest} = substr($payload,($counter * $self->{frame_bytes}));
		}
	else	
		{ 
		$self->{packet_rest} ="";
		}
	
	
	
	
	}

sub flush_audio_frames
	{
	my $self = shift;
	
	my $FH = $self->{outfile};
	my $buffered_frameno = scalar(@{$self->{frame_buffer}});
	print "Found $buffered_frameno audio frames in the buffer\n" if $self->{debug};
	print "flushing for stream $self->{streamcode}\n" if $self->{debug};; 
	my $frames_to_flush =  $buffered_frameno;
	
	print "$self->{chunk_end_correction} should be corrected for end\n" if $self->{debug};
	print "$self->{chunk_start_correction} should be corrected for start\n" if $self->{debug};
	if (!($self->{single_frame}))
		{
		print "we can not sync, we have not got a single frame\n";
		exit;
		}
		
	#  the start of the next chunk is synced before the end of the previous chunk is aligned, because the audio frames
	# numbering suffers from inserting frames. By changing frame number (and numbering)  at the end of the buffer 
	# (which is the beginning of the next chunk) first, the sync of the end of the previos chunk (which naturally occurs before the 
	# beginning of the next chunk) can rely on the computed numbering (X frames at frame number Y)
	#  Confused? Mee too ;-)
	
	if ($self->{chunk_start_correction})
		{
		my ($frames_to_add,$position) = split / : /,$self->{chunk_start_correction};
		$position -= $self->{first_frame_in_buffer};
		print "we should add $frames_to_add at $position for start sync\n" if $self->{debug};
		
		if ($frames_to_add < 0)
			{
			my $last_index = $#{$self->{frame_buffer}};
			print "the frames are index from 0 to $last_index\n" if $self->{debug};
			my @firstbit = @{$self->{frame_buffer}}[0 .. ($position -1)];
			print "Found packets before cut: " . scalar(@firstbit) . "\n" if $self->{debug};
			my @lastbit = @{$self->{frame_buffer}}[($position - $frames_to_add -1 ) .. $last_index];
			print "Found packets after cut: " . scalar(@lastbit) . "\n" if $self->{debug};
			my @final = (@firstbit,  @lastbit);
			print "Total frames after correction: " . scalar(@final) . "\n" if $self->{debug};
			@{$self->{frame_buffer}} = @final;
			}
		
		if ($frames_to_add > 0)
			{
			my $last_index = $#{$self->{frame_buffer}};
			print "the frames are index from 0 to $last_index\n" if $self->{debug};
			my @firstbit = @{$self->{frame_buffer}}[0 .. ($position -1)];
			print "Found packets before cut: " . scalar(@firstbit) . "\n" if $self->{debug};
			my @lastbit = @{$self->{frame_buffer}}[($position) .. $last_index];
			print "Found packets after cut: " . scalar(@lastbit) . "\n" if $self->{debug};
			my @addition;
			for (my $i = 0; $i < $frames_to_add; $i++)
				{
				push @addition, $self->{single_frame};
				}
			my @final = (@firstbit, @addition, @lastbit);
			@{$self->{frame_buffer}} = @final;
			print "Total frames after start correction: " . scalar(@final) . "\n" if $self->{debug};
			}
		
		$self->{chunk_start_correction} = 0;
		}
	
	
	if ($self->{chunk_end_correction})
		{
		print "trying to correct the end sync at cut $self->{cutcounter}\n" if $self->{debug};
		my ($frames_to_add,$position) = split / : /,$self->{chunk_end_correction};
		$position -= $self->{first_frame_in_buffer};
		print "we should add $frames_to_add at $position for end sync\n" if $self->{debug};
		print "First frame in buffer is $self->{first_frame_in_buffer}\n" if $self->{debug};
		if ($frames_to_add < 0)
			{
			my $last_index = $#{$self->{frame_buffer}};
			print "the frames are index from 0 to $last_index\n" if $self->{debug};
			my @firstbit = @{$self->{frame_buffer}}[0 .. ($position + $frames_to_add -1)];
			print "Found packets before cut: " . scalar(@firstbit) . "\n" if $self->{debug};
			my @lastbit = @{$self->{frame_buffer}}[($position) .. $last_index];
			print "Found packets after cut: " . scalar(@lastbit) . "\n" if $self->{debug};
			my @final = (@firstbit,  @lastbit);
			print "Total frames after correction: " . scalar(@final) . "\n" if $self->{debug};
			@{$self->{frame_buffer}} = @final;
			}
		if ($frames_to_add > 0)
			{
			my $last_index = $#{$self->{frame_buffer}};
			print "the frames are index from 0 to $last_index\n" if $self->{debug};
			my @firstbit = @{$self->{frame_buffer}}[0 .. ($position -1)];
			print "Found packets before cut: " . scalar(@firstbit) . "\n" if $self->{debug};
			my @lastbit = @{$self->{frame_buffer}}[($position) .. $last_index];
			print "Found packets after cut: " . scalar(@lastbit) . "\n" if $self->{debug};
			my @addition;
			for (my $i = 0; $i < $frames_to_add; $i++)
				{
				push @addition, $self->{single_frame};
				}
			my @final = (@firstbit, @addition, @lastbit);
			@{$self->{frame_buffer}} = @final;
			print "Total frames after start correction: " . scalar(@final) . "\n" if $self->{debug};
			}
		
		$self->{chunk_end_correction} = 0;
		}

	
	print "trying to flush $frames_to_flush packets\n" if $self->{debug};
	my $counter = 0;
	my $dummy = <STDIN> if ($self->{stop});
	while (scalar(@{$self->{frame_buffer}}) > 0)
		{
		my $frame = shift @{$self->{frame_buffer}};
		print $FH $frame;
		$counter++;
		}
	print "Flushed $counter frames\n" if $self->{debug};
	$self->{frames_written} += $counter;
	
	$buffered_frameno = scalar(@{$self->{frame_buffer}});
	print "Now  $buffered_frameno audio frames in the buffer\n" if $self->{debug};
	$self->{first_frame_in_buffer} += $frames_to_flush;  
	print "First frame in buffer is now: $self->{first_frame_in_buffer}\n" if $self->{debug};
	$self->{frames_since_last_flush} = 0;
	return;
	}

sub do_nothing
	{
	my $self = shift;
	}

sub analyse_video_stream
	{
	my $self = shift;
	my $payload = shift;
	my %pes_header = %{shift @_};
	if (!($pes_header{PTS_decimal})) 
		{
		print "OOPS first video packet w/o ts\n";
		my $dummy = <STDIN> if $self->{stop};
		return;
		}
	
	$self->{cuts}[0] = "$self->{frameno}  : $pes_header{PTS_decimal} : -1 : -1 : $self->{streamcode} : 0 ";
	$self->{cut_flag} = 1;
	
	$self->{packet_start} = pack("H8", "00000100");
	my $seqstart = pack("H8", "000001b3");
	if (!($payload =~ /$seqstart.{8}/)) {print "No SeqHeader in the first Frame, exiting\n"; exit}
	$self->{check_frames} = \&check_GOPs;
	my $bitview = unpack("B64", $&);
	print "Matched: $& at $-[0] bitview of the header\n$bitview \n" if $self->{debug};
	#for (my $i = 0;  $i <64; $i += 8)
	#	{
	#	print "Bit starting at $i: " . substr($bitview, $i, 8) ."\n";
	#	}
	my $fps = substr($bitview,28,4);
	print "we got for fps: $fps\n" if $self->{debug};
	$self->{fps} = 25 if ($fps eq "0011"); # FIXME: Other values possible
	$self->{frame_ticks} =  90000 / $self->{fps};
	open $self->{outfile}, ">$self->{outputpath}/$self->{streamcode}.mpv" || die "Can not open file for m2v: $!\n";
	}

sub check_GOPs
	{
	my $self = shift;
	my $payload = shift;
	my %pes_header = %{shift @_};
	my $FH = $self->{outfile};
	
	# First we just grep for the GOP start sequence
	my $gopstart = pack("H8", "000001b8");
	if (! $self->{GOPno})
		{
		my $seqstart = pack("H8", "000001b3");

		$payload =~ /$seqstart/g;
		if (pos($payload)-4 != 0)
			{
			print "The very first GOP did not start at 0, but at " . (pos($payload)-4) ."\n" if $self->{debug};
			$payload = substr($payload, (pos($payload)-4));
			}
		else
			{
			print "Yes, it did start at 0\n" if $self->{debug};
			}
		}
	my @startcodes  = $payload =~ /$gopstart/gm;
	foreach (@startcodes)
		{
		$self->{GOPno}++;
		$self->{frames_in_last_GOP} = $self->{frames_in_GOP};
		$self->{frames_in_GOP} = 0;
		}
	# If this packet had a timestamp and was no GOP start, we discard the ts. No fiddeling with frame order..
	if (($pes_header{PTS_DTS_flags} ne "00") && (scalar(@startcodes) == 0 ))
		{
		my $waste = pop @{$self->{timestamps}};
		}
	@startcodes = $payload =~ /$self->{packet_start}/gm;
	foreach (@startcodes)
		{
		#print "we matched the frame $counter in the PES-packet\n";
		$self->{frameno}++;
		$self->{frames_in_GOP}++;
		}
	# Next thing would be interesting ...
	if(scalar(@startcodes) > 1) 
		{
		print "more the one frame in the packet\n"; 
		my $dummy = <STDIN> if $self->{stop};
		}
		
	print $FH $payload;
	}


sub search_audio_sync
	{
	# I am not too sure whether this is really a good idea, or whether we should just discard these few audio frames ..
	
	my $self = shift;
	my $payload = shift;
	my %pes_header = %{shift @_};
	my $new_offset = 0;
	
	my $startword = $self->{packet_start};
	print "Audio frames should start with " . unpack("H*" ,$startword) ." $startword\n" if $self->{debug};
	while ($payload =~ /\Q$startword/g)
		{
	
		my $potential_start = (pos($payload)-2);
		print "\tFound a  potential audio syncword at $potential_start: " . unpack("H*" ,substr($payload,$potential_start,3)) . "\n" if $self->{debug};
		if (($potential_start + $self->{frame_bytes}) < length($payload))
			{
			print "\twe should have enough payload left to verify...\n" if $self->{debug};
			print "\tchecking " .  unpack("H4",substr($payload,($potential_start + $self->{frame_bytes}),2)) ."\n" if $self->{debug};
		
			if ($self->{packet_start} eq substr($payload,($potential_start + $self->{frame_bytes}),2))
				{
				print "\tfound an additional audio sync in the right distance\n\tAll should be fine\n" if $self->{debug};
				$new_offset = $potential_start;
				print "\tNew attempt will be started at $new_offset\n" if $self->{debug};
				$self->{packet_rest} ="";
				last;
				}
			else
				{
				print "\tNext audio start not found, continuing to check...\n" if $self->{debug};
				}
			}
		else
			{
			print "\tnot enough payload left to verify\n\tGood Luck\n\tSkipping to next packet\n" if $self->{debug};
			last;
			}
		}
	$self->{frameno}-- if ($self->{frameno} > 0);
	
	
	if ($new_offset)
		{
		&{$self->{check_frames}} ($self, substr($payload,$new_offset), \%pes_header);
		}
	else
		{
		print "\tCould not find anything usefull in this packet, contents dropped, maybe next packet?\n" if $self->{debug};
		}
	}


sub search_first_audio_frame
	{
	# we have to do this because there are recordings that do not have a single aligned audio syncword at the packet start within the first 2000 PES Packets...
	
	my $self = shift;
	my $payload = shift;
	my %pes_header = %{shift @_};
	my $new_offset = 0;
	
	my $startword = "fffc";
	my $translation = unpack ("H*", $payload);
	my @syncword_list;
	
	print "length of payload is ". length($payload). "\n" if $self->{debug};
	print "length of translation is ". length($translation). "\n" if $self->{debug};
	
	print "Audio frames should start with " . ($startword) ." \n" if $self->{debug};
	while ($translation =~ /\Q$startword/g)
		{
	
		my $potential_start = pos($translation)-2;
		if (($potential_start % 2) == 1)
			{
			print "Potential start skipped cause not byte aligned\n";
			next;
			}
		$potential_start /=2;
		print "\tFound a potential audio syncword at $potential_start: " . unpack("H4" ,substr($payload,($potential_start-1),2)) . "\n" if $self->{debug};
		
		push @syncword_list, ($potential_start-1);
		}
		
	
	if (@syncword_list)
		{
		print "Found " . (scalar(@syncword_list)) . " potential synwords in the packet\n";
		return @syncword_list;
		
		}
	else
		{	
		print "\tCould not find anything usefull in this packet, contents dropped, maybe next packet?\n" if $self->{debug};
		}
	return
	}



sub analyse_PES_header
	{
	#this takes long and is ugly ....
	my $PES_packet = ${shift @_};
	my $head_tmpl = 'H6 H2 n1 B8 B8 C1 B40';
	
	my $dts_tmpl = 'B40';
	my %pes_header;
	# =
	#(
	#stream_id 	  => "",
	#PES_packet_length => "",
	#eflags1		  => "",
	#eflags2		  => "",
	#header_length	  => "",
	#payload_start	  => "",
	#data_align	  => "",
	#PTS_DTS_flags	  => "",
	#PTS_raw_bits	  => "",
	#PTS_value_bits	  => "",
	#PTS_decimal	  => ""
	#);

	(my $startcode, $pes_header{stream_id}, 
	    $pes_header{PES_packet_length}, $pes_header{eflags1}, 
	    $pes_header{eflags2}, $pes_header{header_length},
	    $pes_header{PTS_raw_bits}) 
	    = unpack ( $head_tmpl, $PES_packet );
	if ($startcode ne "000001")
		{
		print "No 0x000001 found at current packet\n"; exit
		}
	my $decimal_code = hex($pes_header{stream_id});
	if (((191 < $decimal_code ) &&  (240 > $decimal_code)) ||  ($decimal_code == 189))
		{
		# MPEG2 Audio and Video must have an extended Header as well as AC3
		$pes_header{payload_start} = (9 + $pes_header{header_length});
		# We check whether a TimeStamp is present:
		$pes_header{PTS_DTS_flags} = substr($pes_header{eflags2}, 0, 2);

		#FIXME: overflow of PTS not checked
		if (($pes_header{PTS_DTS_flags} eq "10") || ($pes_header{PTS_DTS_flags} eq "11"))
			{
			$pes_header{PTS_value_bits} = substr($pes_header{PTS_raw_bits},4,3) . substr($pes_header{PTS_raw_bits},8,15) . substr($pes_header{PTS_raw_bits},24,15);
			# decode the timestamp
			$pes_header{PTS_decimal} = unpack("N", (pack ("B32", substr($pes_header{PTS_value_bits},1))));
			$pes_header{PTS_decimal} += 4294967296 if (substr($pes_header{PTS_value_bits},0,1) == 1);
			}
		$pes_header{data_align} = substr($pes_header{eflags1},6,1);
		}
	return %pes_header;
	}


sub print_stats
	{
	my $self = shift;

	foreach (keys(%{$self->{streams}}))
		{
		print "$self->{streams}{$_}{streamtype} for stream $_\n" if $self->{debug};
		if ($self->{ignore_streams}{$_}) 
			{
			print "Ignoring stream $_\n";
			next;
			}
		
		
		if (($self->{streams}{$_}{streamtype} =~ /audio/) || ($self->{streams}{$_}{streamtype} =~ /Layer/))
			{
			print "\naudio stream $_ info ($self->{streams}{$_}{streamtype}):\n";
			print "Sample frequency:    $self->{streams}{$_}->{freq}\n";			# only for Audio 
			print "Bitrate:             $self->{streams}{$_}->{bitrate}\n";     		# only for Audio
			print "Mode:                $self->{streams}{$_}->{mode}\n";     		# only for Audio
			print "Copyright:           $self->{streams}{$_}->{copyright}\n" if  $self->{streams}{$_}->{copyright};		# only for Audio
			print "Frame length (bytes) $self->{streams}{$_}->{frame_bytes}\n";		# only for Audio
			print "Frame length (ticks) $self->{streams}{$_}->{frame_ticks} (90000 / sec)\n\n\n";		# one tick is 1 / 90000 of a second
			}
		if ($self->{streams}{$_}{streamtype} eq "video")
			{
			print "video stream $_ info:\n";
			print "Frame length (ticks) $self->{streams}{$_}->{frame_ticks} (90000 / sec)\n\n\n";		# one tick is 1 / 90000 of a second
			}


		}
	}
	


sub bitrate
	{
	my $self = shift;
	$self->{bitrate} = shift if @_;
	return $self->{bitrate};
	}

sub streamtype
	{
	my $self = shift;
	$self->{streamtype} = shift if @_;
	return $self->{streamtype};
	}

sub streamcode
	{
	my $self = shift;
	$self->{streamcode} = shift if @_;
	return $self->{streamcode};
	}

sub analyse_audio_stream
	{
	# lots of boring typing here ....
	my %bitrates =
	(
	MPEG1_Layer_1 => {
		"0001"	=> "32",
		"0010"	=> "64",
		"0011"	=> "96",
		"0100"	=> "128",
		"0101"	=> "160",
		"0110"	=> "192",
		"0111"	=> "224",
		"1000"	=> "256",
		"1001"	=> "288",
		"1010"	=> "320",
		"1011"	=> "352",
		"1100"	=> "384",
		"1101"	=> "416",
		"1110"	=> "448",
		},
	MPEG1_Layer_2 => {
		"0001"	=> "32",
		"0010"	=> "48",
		"0011"	=> "56",
		"0100"	=> "64",
		"0101"	=> "80",
		"0110"	=> "96",
		"0111"	=> "112",
		"1000"	=> "128",
		"1001"	=> "160",
		"1010"	=> "192",
		"1011"	=> "224",
		"1100"	=> "256",
		"1101"	=> "320",
		"1110"	=> "384",
		},
	MPEG1_Layer_3 => {
		"0001"	=> "32",
		"0010"	=> "40",
		"0011"	=> "48",
		"0100"	=> "56",
		"0101"	=> "64",
		"0110"	=> "80",
		"0111"	=> "96",
		"1000"	=> "112",
		"1001"	=> "128",
		"1010"	=> "160",
		"1011"	=> "192",
		"1100"	=> "224",
		"1101"	=> "256",
		"1110"	=> "320",
		},
	MPEG2_Layer_1 => {
		"0001"	=> "32",
		"0010"	=> "64",
		"0011"	=> "96",
		"0100"	=> "128",
		"0101"	=> "160",
		"0110"	=> "192",
		"0111"	=> "224",
		"1000"	=> "256",
		"1001"	=> "288",
		"1010"	=> "320",
		"1011"	=> "352",
		"1100"	=> "384",
		"1101"	=> "416",
		"1110"	=> "448",
		},
	MPEG2_Layer_2 => {
		"0001"	=> "32",
		"0010"	=> "48",
		"0011"	=> "56",
		"0100"	=> "64",
		"0101"	=> "80",
		"0110"	=> "96",
		"0111"	=> "112",
		"1000"	=> "128",
		"1001"	=> "160",
		"1010"	=> "192",
		"1011"	=> "224",
		"1100"	=> "256",
		"1101"	=> "320",
		"1110"	=> "384",
		},
	MPEG2_Layer_3 => {
		"0001"	=> "8",
		"0010"	=> "16",
		"0011"	=> "24",
		"0100"	=> "32",
		"0101"	=> "64",
		"0110"	=> "80",
		"0111"	=> "56",
		"1000"	=> "64",
		"1001"	=> "128",
		"1010"	=> "160",
		"1011"	=> "112",
		"1100"	=> "128",
		"1101"	=> "256",
		"1110"	=> "320",
		},
	);
	my %freqs =
		(
		MPEG1 => {
			"00" =>  "44100",
			"01" =>  "48000",
			"10" =>  "32000",
			},
		MPEG2 => {
			"00" => "22050",
			"01" => "24000",
			"10" => "16000",
			},
		);

	my $self = shift;
	my $payload = shift;
	my %pes_header = %{shift @_};
	if (!($pes_header{PTS_decimal})) 
		{
		print "OOOPS first packet of audio stream $self->{streamcode} w/o ts\n";
		my $dummy = <STDIN> if $self->{stop};
		return;
		}
	$self->{cuts}[0] = "$self->{frameno}  : $pes_header{PTS_decimal} : -1 : -1 : $self->{streamcode} : ";
	$self->{cut_flag} = 1;
	
	
	my $BITS = unpack("B32", $payload);
	my $streamtype = "";
	
	if (substr($BITS,0,12) ne "111111111111")
		{
		print "No Audio syncword at the beginning of the payload, searching within the packet\n" if $self->{debug};
		my $dummy = <STDIN> if $self->{stop};
		my @syncwordlist = search_first_audio_frame($self, $payload, \%pes_header);
		print "Got " . (scalar(@syncwordlist)) . " potential audio frame start positions back\n" if $self->{debug};
		$payload = substr($payload, $syncwordlist[0]);
		print "new potential payload has a length of " . (length($payload)) . "\n" if $self->{debug};
		$BITS = unpack("B32", $payload)
		}
	$self->{packet_start} = substr($payload,0,2);

	if (substr($BITS,12,1) == 1) {$streamtype = "MPEG1"} else {$streamtype = "MPEG2"}

	if (substr($BITS,13,2) eq "01") {$streamtype .= "_Layer_3"}
	elsif (substr($BITS,13,2) eq "10") {$streamtype .= "_Layer_2"}
	elsif (substr($BITS,13,2) eq "11") {$streamtype .= "_Layer_1"}
	else {print "Streamtype invalid, exiting\n"; exit}
	$self->{copyright} = substr($BITS,15,1);
	$self->{bitrate} = $bitrates{$streamtype}{substr($BITS,16,4)} . "000";
	$self->{freq} = $freqs{substr($streamtype,0,5)}{substr($BITS,20,2)};
	$self->{padding_bit} = substr($BITS,22,1);

	$self->{streamtype} = $streamtype;
	
	if    (substr($BITS,24,2) eq "00") {$self->{mode} = "stereo"}
	elsif (substr($BITS,24,2) eq "01") {$self->{mode} = "joint stereo"}
	elsif (substr($BITS,24,2) eq "10") {$self->{mode} = "dual channel"}
	elsif (substr($BITS,24,2) eq "11") {$self->{mode} = "mono"}
	
	print "$BITS\n Padding: $self->{padding_bit} Mode: $self->{mode}\n" if $self->{debug};
	
	$self->{frame_bytes} = (144 * $self->{bitrate}  / $self->{freq}) + $self->{padding_bit}; #/ for the broken syntax highlighter
	$self->{frame_ticks} = 8 * $self->{frame_bytes} * 90000 / $self->{bitrate}; #/ for the broken syntax highlighter 
	print "streamtype is: $self->{streamtype}\nbitrate is $self->{bitrate}\nfreq is $self->{freq}\ncopyright is $self->{copyright}\nframelength in byte is $self->{frame_bytes}\nin sec: " . ($self->{frame_ticks}/90000) . "\n" if $self->{debug};#/
	print "The start of the packet is: " . substr($BITS, 12) . "\n" if $self->{debug};
	if ($self->{streamtype} eq "MPEG1_Layer_2")
		{
		$self->{check_frames} = \&check_mp2_frames;
		}
	

	open $self->{outfile}, ">$self->{outputpath}/$self->{streamcode}.mpa" || die "Can not open file for mpa: $!\n";
}


sub process_PES_packets
	{
	my $self = shift;
	my $limit = shift;
	my $offset = 0;  		# stores the position we are at in the current buffer
	my $plength = 0;
	$self->{IFH} = undef; 		# Stores the filehandle for the current file
	$self->{EOF} = 0; 		# All files processed ?
	$self->{total_input} = 0;
	my $lengthcounter = 0;
	my $packetcounter = 0;
	
	
		
	my $buffer = readNextChunk($self); 	# First we read the first chunk of data
	
	while (($self->{EOF} eq 0 || $offset + $plength + 1) < (length($buffer)))
		{
		# there are at leat six bytes of header at the beginning of a PES packet
		(my $startcode, my $streamcode, my $plength) = unpack ( "H6 H2 n1", substr($buffer, $offset, 6) );
		if ($startcode ne "000001")
			{
			print "No 0x000001 found at current packet\n$startcode"; exit
			}
		if ((($offset + $plength + 150) > (length($buffer))) && ($self->{EOF} == 0))
			{
			my $helpbuffer = substr($buffer, $offset);
			$buffer = $helpbuffer . readNextChunk($self);
			$offset = 0;
			}

		my $packet = substr($buffer, $offset, ($plength + 6));
		$offset +=  $plength + 6;
		$packetcounter++;
		
		
		if ($self->{ignore_streams}{$streamcode})
			{
			print "ignoring packet for stream $streamcode\n" if $self->{debug};
			next;
			}
		
		if (!($self->{streams}{$streamcode}))
			{
			print "New Stream with id $streamcode detected. Ignoring the stream\n";
			$self->{ingnore_streams}{$streamcode} = 1;
			next;
			}
		
		if ($self->{dump_packets})
			{
			print "Dumping packet of stream $streamcode ($self->{dump_packets} to dump left)\n";
			my $DUMPFH;
			if (! $self->{"$streamcode.pes_dump"})
				{
				open $DUMPFH, ">$streamcode.pes_dump" || die "Can not open dumpfile: $!\n";
				$self->{"$streamcode.pes_dump"} = $DUMPFH;
				}
			$DUMPFH = $self->{"$streamcode.pes_dump"};
			print $DUMPFH $packet;
			if ($self->{dump_packets}-- == 1)
				{
				exit;
				}
			next;
			}
		
		if ($self->{streams}{$streamcode}->{cut_flag} == 1)
			{

			$self->{streams}{$streamcode}->{cut_flag} = 0;
			#print "cut detected in stream $streamcode at frame $self->{streams}{$streamcode}->{frameno}, resetting packet_before_sync counter\n";
			
			# We can not start to sync right after we detected the first cut
			# We need some info about the other streams as well
			# So we wait for at least 300 PES-Packets before we try to compute the desync
			# If we detect a cut in the next stream, the timer is set to 300 again
			#Thereby we make sure that we have enough audio frames of each stream to sync
			$self->{packets_before_sync} = 300;
			}
		
		
		# we do not want to take too much memory for buffered audio frames
		if (($self->{streams}{$streamcode}->{frames_since_last_flush} > 300) &&  ($self->{packets_before_sync} == 0))
			{
			print "going to flush a bit ...\n" if $self->{debug};
			$self->{streams}{$streamcode}->flush_audio_frames($self->{streams}{$streamcode});
			}
		if ($self->{packets_before_sync}  > 0)
			{
			$self->{packets_before_sync}--;
			if ($self->{packets_before_sync} == 0)
				{
				compute_desync($self);
				}
			}
		
		$self->{streams}{$streamcode}->add_PES_packet(\$packet);
		
		}
	
	print "$packetcounter PES packets processed\n";
	
	
	foreach (keys (%{$self->{streams}}))
		{
		print "Final flush for stream $_\n";
		
		if ($_ eq $self->{masterstream}) 
			{
			$self->{streams}{$_}{frames_written} = $self->{streams}{$_}{frameno};
			}
		
		if ($self->{ignore_streams}{$_})
			{
			print "ignoring packet for stream $_ \n" if $self->{debug};
			next;
			}
		my $FH = $self->{streams}{$_}{outfile};
		if (!($FH)) {print "Could not get the File Handle\n";exit}
		
		while (scalar(@{$self->{streams}{$_}{frame_buffer}}) > 0 )
			{
			my $frame = shift @{$self->{streams}{$_}{frame_buffer}};
			print "Still " . scalar(@{$self->{streams}{$_}{frame_buffer}}) ." frames to print\n" if ($self->{debug});
			print $FH $frame;
			$self->{streams}{$_}{frames_written}++;
			}
				
		my $seconds = $self->{streams}{$_}{frames_written} * $self->{streams}{$_}{frame_ticks} / 90000; #/
		my $tsno = scalar(@{$self->{streams}{$_}{timestamps}});
		print "$self->{streams}{$_}{frames_written} frames written for stream $_ ($seconds sec) $tsno ts found\n";
		}
	
	}

sub init_PES_stream
	{
	my $self = shift;
	my $limit = shift;
	my $offset = 0;  		# stores the position we are at in the current buffer
	my $plength = 0;
	$self->{IFH} = undef; 		# Stores the filehandle for the current file
	$self->{EOF} = 0; 		# All files processed ?
	$self->{total_input} = 0;
	
	my $packetcounter = 0;
	
	print "Initialising and analysing the streams....\n";
	my @save_file_list = @{$self->{files}};
	
	my $buffer = readNextChunk($self); 	# First we read the first chunk of data
	
	while ((($self->{EOF} eq 0 || $offset + $plength + 1) < (length($buffer))) && ($packetcounter < 2000))
		{
		# there are at leat six bytes of header at the beginning of a PES packet
		(my $startcode, my $streamcode, my $plength) = unpack ( "H6 H2 n1", substr($buffer, $offset, 6) );
		if ($startcode ne "000001")
			{
			print "No 0x000001 found at current packet\n$startcode"; exit
			}
		if ((($offset + $plength + 150) > (length($buffer))) && ($self->{EOF} == 0))
			{
			my $helpbuffer = substr($buffer, $offset);
			$buffer = $helpbuffer . readNextChunk($self);
			$offset = 0;
			}

		my $packet = substr($buffer, $offset, ($plength + 6));
		$offset +=  $plength + 6;
		$packetcounter++;
		print "analysed the first $packetcounter packets...\n" if $self->{debug};
		if (!$self->{streams}{$streamcode}) 
			{
			if ($self->{ignore_streams}{$streamcode})
				{
				next;
				}
			my $decimal_code = hex($streamcode);
			if ((223 < $decimal_code ) &&  (240 > $decimal_code))
				{
				if ($self->{masterstream})
					{
					print "Video stream already defined, but found stream $streamcode in addition. Ingnoring stream $streamcode\n";
					$self->{ingnore_streams}{$streamcode} = 1;
					next;
					}
				else
					{
					$self->{streams}{$streamcode} = MPEGSTREAM->new
							(
							streamcode 	=> "$streamcode",
							outputpath 	=> $self->{outputpath}, 
							masterstream 	=> $streamcode, 
							debug 		=> $self->{debug},
							dump_payload	=> $self->{dump_payload},
							);
					$self->{masterstream} = $streamcode;
					print "Created new MPEG stream object for stream $streamcode, master video stream\n";
					}
				}
			else
				{
				$self->{streams}{$streamcode} = MPEGSTREAM->new
							(
							streamcode 	=> "$streamcode", 
							masterstream 	=> $self->{masterstream}, 
							outputpath 	=> $self->{outputpath}, 
							debug 		=> $self->{debug},
							dump_payload	=> $self->{dump_payload}
							);	
				print "Created new MPEG stream object for stream $streamcode \n";
				}
						
			}
		my %pes_header = analyse_PES_header(\$packet);
		if (!($self->{streams}{$streamcode}{check_frames}))
			{
			print "sending frame number $packetcounter for analysis of stream $streamcode\n" if $self->{debug};
			init_stream($self->{streams}{$streamcode}, substr($packet, $pes_header{payload_start}), \%pes_header);
			}
		}
		
		
	@{$self->{files}} = @save_file_list;
	print "analysed the first $packetcounter packets...\n";
	my $IFH = $self->{IFH};
	close $IFH;
	
	if (!($self->{masterstream}))
		{
		die "No video stream could be found within the first 2000 packets, exiting\n" if (! $self->{ignore_streams});
		}
	
	foreach (keys(%{$self->{streams}}))
		{
		$self->{streams}{$_}{masterstream} = $self->{masterstream};
		
		if ((!($self->{streams}{$_}{check_frames})) && (!($self->{ignore_streams}{$_})))
			{
			print "The contents of stream $_ could not be identified, the stream will be skipped!\n";
			$self->{ignore_streams}{$_} = 1;
			}
		else
			{
			#$self->{ignore_streams}{$_} = 0;
			}
		}
	
	print_stats($self);
	my $dummy = <STDIN> if $self->{stop};  
	return;
	}


sub readNextChunk
{
	my $self = shift;
	my $IFH = $self->{IFH};
	

	if (!($self->{IFH}))
		{
		
		my $firstfile = shift @{$self->{files}};
		print "trying to open $firstfile\n" if $self->{debug};
		open $IFH, $firstfile  || die "$! happend while opening $firstfile\n";
		$self->{IFH} = $IFH;
		}
	
	my $rbuffer;
	
	my $rbytes = sysread $IFH, $rbuffer, 10000000, 0;

	
	if ($rbytes != 10000000 )
		{
			if ((scalar(@{$self->{files}}))== 0)
			{
			print "all Input files processed\n";
			print "EOF reached\n";
			$self->{EOF} = 1;
			}
		else
			{
			my $nextf = shift @{$self->{files}};
			print "$nextf is the next file\n";
			close $IFH;
			open $IFH, $nextf  || die "$! happend while opening $nextf\n";
			$self->{IFH} = $IFH;
			my $helpbuffer;
			my $helprbytes = sysread $self->{IFH}, $helpbuffer, (10000000 - length($rbuffer)), 0;
			$rbuffer .= $helpbuffer;
			$rbytes += $helprbytes;
			}
		}
 	$self->{total_input} += $rbytes;
	my $status = sprintf ("%4d",  ($self->{total_input} / 1000000));#/
	print "$status Mbytes read\n"; #if $self->{debug};  f
	return ($rbuffer);
}

